%!octave 
%
%
%
%       OS : GNU/Linux 4.2.5
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: w2v_fit_cbow.m,v 1.8 2017/06/05 13:40:01 mechanoid Exp $

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function [Vi0,Vo0,er] = w2v_fit_cbow(C,W,Vi,Vo),
 
   s = size(C);
   Cavg = reshape( sum(C,2)./(s(2)*2), [s(1),s(3)] ) ;	 

   a = 0.91; % скорость обучения

   nn = 1000 ;
   er = [ 1e10 ] ; % история изменения ошибки
   for i=1:nn,
      [H,U,O] = w2v_step(Cavg,Vi,Vo) ; % вычисляем состояния слоёв и выход

      er(end+1) = w2v_loss(U,W) ; % вычисляем ошибку 
      printf("[i] эпоха %i/%i, ошибка обучения: %f     \r",i,nn, er(end)) ;
      if( er(end-1) < er(end) ) break; endif; % если ошибка растёт то останавливаем обучение 

      [gVi,gVo] = w2v_grad(C,W,H,O,Vo) ; % вычисляем градиент ф-ции потери

      [gVi,gVo] = w2v_weigth_norm(gVi,gVo) ; % нормируем значения градиента
    
      Vi0 = Vi ; 
      Vo0 = Vo ; 

      [Vi,Vo] = w2v_weigth_add(Vi,Vo,gVi,gVo,-a) ; % корректируем веса
   endfor

   er(1) = [] ;
   er(end) = [] ;
   printf("[i] эпоха %i/%i, ошибка обучения: %f     \n",i,nn, er(end)) ;

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% вычисляем градиент ф-ции потери
% методом cbow
%
function [gVi,gVo] = w2v_grad(C,W,H,O,Vo),
   s = size(C) ; 
   cw = s(2) ; % размер окна контекста 

   D = O - W ; 
   gVo = transpose(H) * D ; 

   gVi = 0.0 ;
   for i=1:cw,
      Ci = reshape( C(:,i,:), s([1,3]) ) ; % слово i контекста
      gVi += transpose(Ci) * D * transpose(Vo) ; 
   endfor

  
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% ф-ция потери для cbow
%
function retval = w2v_loss(U,W),
   n = size(W,1); % количество примеров
   Us = sum(log(sum(exp(U),2))) ;
   Uo = sum(sum(U.*W,2)) ; % состояния выходного слоя соответвующие слову y
   retval = abs( Us - Uo ) ./ n ;

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function [ H,U,O ] = w2v_step(C,Vi,Vo),
   H = C * Vi ; % значения скрытого слоя
   U = H * Vo ; % состояния выходного слоя
   O = softmax(U) ; % выход сети
endfunction


