%!octave 
%
%
%
%       OS : GNU/Linux 4.2.5
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: w2v_fit.m,v 1.2 2017/06/05 12:49:02 mechanoid Exp mechanoid $


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function [Vi0,Vo0,er] = w2v_fit_skip_gram(C,W,Vi,Vo),
   a = 0.91; % скорость обучения

   nn = 1000 ;
   er = [ 1e10 ] ; % история изменения ошибки
   for i=1:nn,
      [H,U,O] = w2v_step(W,Vi,Vo) ; % вычисляем состояния слоёв и выход

      er(end+1) = w2v_loss(U,C) ; % вычисляем ошибку 
      printf("[i] эпоха %i/%i, ошибка обучения: %f     \r",i,nn, er(end)) ;

      if( er(end-1) < er(end) ) break; endif; % если ошибка растёт то останавливаем обучение 

      [gVi,gVo] = w2v_grad(C,W,H,O,Vo) ; % вычисляем градиент ф-ции потери

      [gVi,gVo] = w2v_weigth_norm(gVi,gVo) ; % нормируем значения градиента

      Vi0 = Vi ; 
      Vo0 = Vo ; 

      [Vi,Vo] = w2v_weigth_add(Vi,Vo,gVi,gVo,-a) ; % корректируем веса
   endfor

   er(1) = [] ;
   er(end) = [] ;

   printf("[i] эпоха %i/%i, ошибка обучения: %f     \n",i,nn, er(end)) ;

endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function [ H,U,O ] = w2v_step(W,Vi,Vo),
   H = W * Vi ; % значения скрытого слоя
   U = H * Vo ; % состояния выходного слоя
   O = softmax(U) ; % выход сети
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% ф-ция потери для skip-gram
%
function retval = w2v_loss(U,C),
   s = size(C) ; % размер набора слов контекста
   n = s(1) ; % количество примеров
   v = s(3) ; % количество слов в словаре
   cws = s(2) ; % размер окна контекста 

   Us = sum(log(sum(exp(U),2))) ;
   
   Uo = 0.0;
   for i=1:cws, % для всех слов контекста
      Ci = reshape(C(:,i,:),[n,v]) ; % набор слов контекста i
      Uo += sum(sum(U.*Ci,2)) ; % значения выходного слоя для слов x контекста i
   endfor

   retval = abs(Us - Uo) ./ n ;

  
endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% вычисляем градиент ф-ции потери
% методом skip-gram
%
function [gVi,gVo] = w2v_grad(C,W,H,O,Vo),
   s = size(C) ;  % размеры набора слов контекста
   cws = s(2) ; % размер окна контекста 
   
   sw = size(Vo) ; % размеры матрицы связей выходного слоя

   gVi = gVo = 0.0 ; 

   for i=1:cws,
      Ci = reshape( C(:,i,:), s([1,3]) ) ; % слово i контекста
      D = O - Ci ; % ошибка на слове контекста i
      gVo += transpose(H) * D ;
      gVi += transpose(D) * W * transpose(Vo) ; 
   endfor

endfunction


